use rocket::{
	http::{Cookie, CookieJar},
	time::Duration,
};
use totp_rs::{Secret, TOTP};

const ISSUER: &'static str = "dolltags";

fn make_cookie<'a>(value: String) -> Cookie<'a> {
	Cookie::build(("otp_init", value))
		.max_age(Duration::minutes(5))
		.http_only(true)
		.same_site(rocket::http::SameSite::Strict)
		.path("/account/settings")
		.build()
}

/// Creates a TOTP instance using in-code defaults and user OTP settings.
/// i use the internal user account UUID instead of their username to cover the case
/// where a user needs to change their account username.
pub fn make_totp(account_id: &str, secret: Vec<u8>) -> Result<TOTP, totp_rs::TotpUrlError> {
	TOTP::new(
		totp_rs::Algorithm::SHA1,
		6,
		1,
		30,
		secret,
		Some(String::from(ISSUER)),
		String::from(account_id),
	)
}

/// Adds the provided OTP secret as encrypted cookie for 5 minutes
pub fn cache_secret<'a>(cookies: &CookieJar<'a>, secret: &'a Secret) {
	if let Secret::Encoded(b32) = secret.to_encoded() {
		cookies.add_private(make_cookie(b32));
	}
}

/// Tries to get a potentially existing secret from the cookies.
/// If no cookie is set or if the cookie's value doesn't parse to
/// a valid base32 secret this returns None
pub fn get_secret(cookies: &CookieJar<'_>) -> Option<Secret> {
	cookies.get_private("otp_init").and_then(|cookie| {
		let val = cookie.value_trimmed();
		Secret::Encoded(String::from(val)).to_raw().ok()
	})
}

/// Manually revokes the encrypted cookie
pub fn remove_secret(cookies: &CookieJar<'_>) {
	cookies.remove_private(make_cookie(String::from("woof")))
}
