use std::net::IpAddr;

use rocket::{form::Form, http::CookieJar, response::Redirect, tokio::task};
use rocket_dyn_templates::{context, Template};
use totp_rs::Secret;

use crate::{
	auth::{self, pw},
	db::{
		self,
		otp::METHOD_TOTP,
		schema::{DollTagsDb, User},
	},
	ids::generate_recovery_key,
	pages::CommonTemplateState,
	routes::{self, error_handlers::PageResult},
};

#[get("/settings/totp?<invalid_code>")]
pub async fn show_totp_enable_start(
	mut db: DollTagsDb,
	invalid_code: bool,
	user: User,
	cookies: &CookieJar<'_>,
	meta: CommonTemplateState,
) -> PageResult {
	if db::otp::has_otp(&mut *db, &user.id, METHOD_TOTP).await? {
		return Ok(Redirect::to(uri!("/account", routes::account::settings::show_settings)).into());
	}

	// cookie recovery in case it was a form error and not a new session
	let mut totp_secret = auth::otp::get_secret(cookies);
	if totp_secret.is_none() {
		let new_secret = Secret::generate_secret();
		auth::otp::cache_secret(cookies, &new_secret);
		totp_secret = Some(new_secret);
	}
	let totp_secret = totp_secret.unwrap();

	let totp = auth::otp::make_totp(&user.id.to_string(), totp_secret.to_bytes()?)?;
	let totp_qrcode = totp.get_qr_base64()?;

	Ok(Template::render(
		"account/otp/start",
		context! {
			meta,
			totp_qrcode,
			invalid_code,
			secret: totp.get_secret_base32(),
		},
	)
	.into())
}

#[derive(FromForm)]
pub struct OtpEnableForm {
	#[field(validate = len(6..=6))]
	pub otp_code: String,
}

#[post("/settings/totp", data = "<form>")]
pub async fn handle_totp_enable_start(
	mut db: DollTagsDb,
	form: Form<OtpEnableForm>,
	user: User,
	cookies: &CookieJar<'_>,
	meta: CommonTemplateState,
	client_ip: IpAddr,
) -> PageResult {
	if db::otp::has_otp(&mut *db, &user.id, METHOD_TOTP).await? {
		return Ok(Redirect::to(uri!("/account", routes::account::settings::show_settings)).into());
	}
	let secret = match auth::otp::get_secret(cookies) {
		Some(v) => v,
		None => return Ok(Redirect::to(uri!("/account", show_totp_enable_start(false))).into()),
	};

	let totp = auth::otp::make_totp(&user.id.to_string(), secret.to_bytes()?)?;

	if !totp.check_current(&form.otp_code)? {
		return Ok(Redirect::to(uri!("/account", show_totp_enable_start(true))).into());
	}

	warn!(
		"[audit|{}] [{}] enabled TOTP 2FA",
		client_ip,
		user.id.to_string(),
	);

	let recovery_key = generate_recovery_key();

	{
		let recovery_key = recovery_key.clone();
		let hashed_recovery_key = task::spawn_blocking(move || pw::hash(&recovery_key)).await??;

		db::otp::add_otp_method(
			&mut *db,
			&user.id,
			&totp.get_secret_base32(),
			&hashed_recovery_key,
		)
		.await?;
	}

	auth::otp::remove_secret(cookies);

	Ok(Template::render(
		"account/otp/confirm",
		context! {
			meta,
			recovery_key,
		},
	)
	.into())
}

#[get("/settings/totp/generate-key")]
pub async fn show_confirm_totp_regenerate_key(
	mut db: DollTagsDb,
	user: User,
	meta: CommonTemplateState,
) -> PageResult {
	if !db::otp::has_otp(&mut *db, &user.id, METHOD_TOTP).await? {
		return Ok(Redirect::to(uri!("/account", routes::account::settings::show_settings)).into());
	}

	Ok(Template::render(
		"account/otp/confirm_regenerate_key",
		context! {
			meta,
		},
	)
	.into())
}

#[post("/settings/totp/generate-key")]
pub async fn regenerate_key(
	mut db: DollTagsDb,
	user: User,
	meta: CommonTemplateState,
	client_ip: IpAddr,
) -> PageResult {
	if !db::otp::has_otp(&mut *db, &user.id, METHOD_TOTP).await? {
		return Ok(Redirect::to(uri!("/account", routes::account::settings::show_settings)).into());
	}

	warn!(
		"[audit|{}] [{}] regenerated TOTP 2FA recovery key",
		client_ip,
		user.id.to_string(),
	);

	let recovery_key = generate_recovery_key();

	{
		let recovery_key = recovery_key.clone();
		let hashed_recovery_key = task::spawn_blocking(move || pw::hash(&recovery_key)).await??;

		db::otp::change_recovery_key(&mut *db, &user.id, METHOD_TOTP, &hashed_recovery_key).await?;
	}

	Ok(Template::render(
		"account/otp/regenerate_key",
		context! {
			recovery_key,
			meta,
		},
	)
	.into())
}

#[get("/settings/totp/disable")]
pub async fn show_confirm_totp_disable(
	mut db: DollTagsDb,
	user: User,
	meta: CommonTemplateState,
) -> PageResult {
	if !db::otp::has_otp(&mut *db, &user.id, METHOD_TOTP).await? {
		return Ok(Redirect::to(uri!("/account", routes::account::settings::show_settings)).into());
	}

	Ok(Template::render(
		"account/otp/disable",
		context! {
			meta,
		},
	)
	.into())
}

#[post("/settings/totp/disable")]
pub async fn handle_confirm_totp_disable(
	mut db: DollTagsDb,
	user: User,
	client_ip: IpAddr,
) -> PageResult {
	if !db::otp::has_otp(&mut *db, &user.id, METHOD_TOTP).await? {
		return Ok(Redirect::to(uri!("/account", routes::account::settings::show_settings)).into());
	}

	warn!(
		"[audit|{}] [{}] deactivated TOTP 2FA",
		client_ip,
		user.id.to_string(),
	);

	db::otp::delete_otp_method(&mut *db, &user.id, METHOD_TOTP).await?;

	Ok(Redirect::to(uri!("/account", routes::account::settings::show_settings)).into())
}
