use std::net::IpAddr;

use regex::Regex;
use rocket::{
	form::{self, Contextual, Form},
	http::CookieJar,
	response::Redirect,
	tokio::task,
};
use rocket_dyn_templates::{context, Template};
use totp_rs::Secret;

use crate::{
	auth::{
		otp::make_totp,
		pw,
		session::{self, AuthSession},
	},
	db::{
		otp::{self, METHOD_TOTP},
		schema::{DollTagsDb, User},
		user,
	},
	pages::CommonTemplateState,
	routes::{
		self,
		error_handlers::{PageResponse, PageResult},
	},
};

#[derive(Debug, FromForm)]
pub struct AuthForm<'a> {
	#[field(validate=len(2..=256))]
	pub username: &'a str,
	#[field(validate=len(8..))]
	pub password: &'a str,
}

// needed cuz the var. name changes how the query string is parsed
#[allow(unused_variables)]
#[get("/login?<next>")]
pub fn show_login(
	next: Option<&str>,
	maybe_loggedin: Option<User>,
	meta: CommonTemplateState,
) -> PageResult {
	if maybe_loggedin.is_some() {
		let next = String::from(next.unwrap_or("/"));
		Ok(Redirect::to(next).into())
	} else {
		Ok(Template::render("account/login", context! {meta}).into())
	}
}

#[post("/login?<next>", data = "<form>")]
pub async fn handle_login(
	mut db: DollTagsDb,
	next: Option<&str>,
	form: Form<Contextual<'_, AuthForm<'_>>>,
	cookies: &CookieJar<'_>,
	maybe_loggedin: Option<User>,
	meta: CommonTemplateState,
	client_ip: IpAddr,
) -> PageResult {
	if maybe_loggedin.is_some() {
		let next = String::from(next.unwrap_or("/"));
		return Ok(Redirect::to(next).into());
	}

	let values = match &form.value {
		None => {
			return Ok(Template::render("account/login", context! {failure: true, meta}).into())
		}
		Some(v) => v,
	};

	let miss = (|username: String| {
		move || {
			warn!("[audit|{}] login failure ({})", client_ip, username);
			PageResponse::Page(Template::render(
				"account/login",
				context! {failure: true, meta},
			))
		}
	})(String::from(values.username));

	warn!("[audit|{}] login attempt ({})", client_ip, &values.username);

	let user_in_db = user::get(&mut *db, &values.username).await?;
	let user = match user_in_db {
		None => {
			task::spawn_blocking(move || pw::verify("meow", "$argon2i$v=19$m=65536,t=3,p=1$fJ+f67UGHB+EIjGIDEwbSQ$V/nZPHmdyqHq8fTBTdt3sEmTyr0W7i/F98EIxaaJJt0")).await??;
			return Ok(miss());
		}
		Some(v) => v,
	};

	let password = String::from(values.password);
	let right_password =
		task::spawn_blocking(move || pw::verify(&password, &user.password)).await??;

	if right_password && user.enabled {
		let enabled_methods = otp::list_enabled_methods(&mut *db, &user.id).await?;

		if enabled_methods.len() == 0 {
			session::login(cookies, &user.id);
			warn!(
				"[audit|{}] [{}] login successful ({})",
				client_ip,
				user.id.to_string(),
				&values.username
			);
			let next_url = String::from(next.unwrap_or("/"));
			Ok(Redirect::to(next_url).into())
		} else {
			warn!(
				"[audit|{}] [{}] login user credentials validated; 2FA needed ({})",
				client_ip,
				user.id.to_string(),
				&values.username
			);

			// only one impl. for now; flow and UI may change if/when more are added
			if enabled_methods[0].contains(METHOD_TOTP) {
				session::init_2fa_auth(cookies, &user.id);

				Ok(Redirect::to(uri!(show_2fa_form(next))).into())
			} else {
				panic!(
					"login - 2FA redirect - {:?} not implemented",
					enabled_methods
				);
			}
		}
	} else {
		Ok(miss())
	}
}

#[get("/login/2fa?<next>")]
pub fn show_2fa_form(
	maybe_loggedin: Option<User>,
	_auth_session: AuthSession,
	cookies: &CookieJar<'_>,
	next: Option<&str>,
	meta: CommonTemplateState,
) -> PageResult {
	if maybe_loggedin.is_some() {
		let next = String::from(next.unwrap_or("/"));
		session::clear_2fa_auth(cookies);
		Ok(Redirect::to(next).into())
	} else {
		Ok(Template::render(
			"account/otp/login",
			context! {
				recovery_url: uri!(show_2fa_recover_form(next)),
				meta,
			},
		)
		.into())
	}
}

#[derive(Debug, FromForm)]
pub struct TOTP2FAForm<'a> {
	#[field(validate=len(6..=6))]
	pub otp_code: &'a str,
}

#[post("/login/2fa?<next>", data = "<form>")]
pub async fn handle_2fa_form(
	mut db: DollTagsDb,
	maybe_loggedin: Option<User>,
	auth_session: AuthSession,
	next: Option<&str>,
	cookies: &CookieJar<'_>,
	form: Form<Contextual<'_, TOTP2FAForm<'_>>>,
	meta: CommonTemplateState,
	client_ip: IpAddr,
) -> PageResult {
	if maybe_loggedin.is_some() {
		let next = String::from(next.unwrap_or("/"));
		return Ok(Redirect::to(next).into());
	}

	let otp_config = otp::get_otp_method(&mut *db, &auth_session.0, METHOD_TOTP).await?;

	let (form_data, otp_config) = match (&form.value, otp_config) {
		(Some(form), Some(otp)) => (form, otp),
		_ => {
			return Ok(Template::render(
				"account/otp/login",
				context! {
					recovery_url: uri!(show_2fa_recover_form(next)),
					failure: true,
					meta,
				},
			)
			.into())
		}
	};

	let totp = make_totp(
		&auth_session.0.to_string(),
		Secret::Encoded(otp_config.secret_seed)
			.to_raw()?
			.to_bytes()?,
	)?;

	if totp.check_current(form_data.otp_code)? {
		session::clear_2fa_auth(cookies);
		session::login(cookies, &auth_session.0);

		warn!(
			"[audit|{}] [{}] TOTP 2FA login successful",
			client_ip,
			&auth_session.0.to_string()
		);
		let next_url = String::from(next.unwrap_or("/"));
		Ok(Redirect::to(next_url).into())
	} else {
		Ok(Template::render(
			"account/otp/login",
			context! {
				recovery_url: uri!(show_2fa_recover_form(next)),
				failure: true,
				meta,
			},
		)
		.into())
	}
}

#[get("/login/2fa/recover?<next>")]
pub fn show_2fa_recover_form(
	maybe_loggedin: Option<User>,
	_auth_session: AuthSession,
	next: Option<&str>,
	meta: CommonTemplateState,
) -> PageResult {
	if maybe_loggedin.is_some() {
		let next = String::from(next.unwrap_or("/"));
		return Ok(Redirect::to(next).into());
	}

	Ok(Template::render("account/otp/recovery_login", context! {meta}).into())
}

#[derive(Debug, FromForm)]
pub struct TOTP2FARecoverForm<'a> {
	#[field(validate=len(16..=16))]
	pub recovery_key: &'a str,
}

#[post("/login/2fa/recover?<next>", data = "<form>")]
pub async fn handle_2fa_recover_form(
	mut db: DollTagsDb,
	maybe_loggedin: Option<User>,
	auth_session: AuthSession,
	cookies: &CookieJar<'_>,
	form: Form<Contextual<'_, TOTP2FARecoverForm<'_>>>,
	next: Option<&str>,
	meta: CommonTemplateState,
	client_ip: IpAddr,
) -> PageResult {
	if maybe_loggedin.is_some() {
		let next = String::from(next.unwrap_or("/"));
		return Ok(Redirect::to(next).into());
	}

	let otp_config = otp::get_otp_method(&mut *db, &auth_session.0, METHOD_TOTP).await?;

	let (form_data, otp_config) = match (&form.value, otp_config) {
		(Some(form), Some(otp)) => (form, otp),
		_ => {
			return Ok(Template::render(
				"account/otp/recovery_login",
				context! {failure: true, meta},
			)
			.into())
		}
	};

	let submitted_key = String::from(form_data.recovery_key);
	let hashed_key = otp_config.recovery_key;
	let right_recovery_key =
		task::spawn_blocking(move || pw::verify(&submitted_key, &hashed_key)).await??;

	if right_recovery_key {
		warn!(
			"[audit|{}] [{}] TOTP 2FA recovery key used - disabling TOTP 2FA",
			client_ip,
			&auth_session.0.to_string()
		);

		otp::delete_otp_method(&mut *db, &auth_session.0, METHOD_TOTP).await?;
		session::clear_2fa_auth(cookies);
		session::login(cookies, &auth_session.0);

		Ok(Redirect::to(uri!(
			"/account",
			routes::account::settings::show_settings,
			"#otp"
		))
		.into())
	} else {
		Ok(Template::render("account/otp/recovery_login", context! {failure: true, meta}).into())
	}
}

#[get("/register")]
pub fn show_register(maybe_loggedin: Option<User>, meta: CommonTemplateState) -> PageResult {
	if maybe_loggedin.is_some() {
		return Ok(Redirect::to("/").into());
	}

	Ok(Template::render(
		"account/register",
		context! {
			previous: form::Context::default(),
			meta
		},
	)
	.into())
}

#[derive(Debug, FromForm)]
pub struct RegisterForm<'a> {
	#[field(validate=len(1..=256))]
	pub username: &'a str,
	#[field(validate=len(8..))]
	pub password: &'a str,
	#[field(validate=validate_email())]
	pub email: &'a str,

	// if it's set to true, a (bad) bot is submitting the form since it's a hidden field
	pub anti_bot: bool,
	// if it isn't set to true, we re-request it to be set
	pub inhuman: bool,
}

fn validate_email<'v>(email: &str) -> form::Result<'v, ()> {
	let email_re = Regex::new(r".+@.+").unwrap();
	if email.len() == 0 {
		return Ok(());
	}

	if email.len() >= 3 && email.len() <= 254 && email_re.is_match(email) {
		Ok(())
	} else {
		Err(form::Error::validation("this doesn't appear to be a valid email").into())
	}
}

#[post("/register", data = "<form>")]
pub async fn handle_register(
	mut db: DollTagsDb,
	form: Form<Contextual<'_, RegisterForm<'_>>>,
	cookies: &CookieJar<'_>,
	maybe_loggedin: Option<User>,
	meta: CommonTemplateState,
	client_ip: IpAddr,
) -> PageResult {
	if maybe_loggedin.is_some() {
		return Ok(Redirect::to("/").into());
	}

	let values = match form.value {
		Some(ref v) => v,
		None => {
			debug!(
				"account registration form invalid, context: {:?}",
				&form.context
			);

			return Ok(Template::render(
				"account/register",
				context! {
					meta,
					previous: &form.context,
				},
			)
			.into());
		}
	};

	let password = String::from(values.password);
	let hashed_password = task::spawn_blocking(move || pw::hash(&password)).await??;

	let account_id = user::create(
		&mut *db,
		values.username,
		&hashed_password,
		if values.email.len() != 0 {
			Some(values.email)
		} else {
			None
		},
	)
	.await?;

	warn!(
		"[audit|{}] [{}] account creation",
		client_ip,
		&account_id.to_string()
	);
	session::login(cookies, &account_id);

	Ok(Redirect::to(uri!(
		"/account",
		crate::routes::form::register_tag::show_register
	))
	.into())
}

#[get("/logout")]
pub fn logout(cookies: &CookieJar<'_>) -> Redirect {
	session::logout(cookies);

	Redirect::to("/")
}
